import re
from collections import (
    Mapping,
)
from typing import (
    Any,
    NewType,
    Optional,
    Union,
)

import asyncpg
from django.apps import (
    apps,
)
from django.db import (
    DEFAULT_DB_ALIAS,
    connections as django_connections,
)
from django.db.models import (
    Field,
    Model,
)


ConnectionAlias = NewType('ConnectionAlias', str)
ConnectionInterface = Union[asyncpg.Pool, asyncpg.Connection]


connections: Mapping[ConnectionAlias, asyncpg.Pool] = {}


async def _make_connection(alias: ConnectionAlias) -> asyncpg.Pool:
    """
    Создаёт пул подключений asyncpg.Pool с теми же параметрами, что
    используются для создания подключения alias в django.
    Для внутреннего использования, используйте публичные интерфейсы
    get_pool и get_connection

    Args:
        alias: псевдоним БД в django

    Returns:
        пул подключений asyncpg.Pool

    """
    dj_connection = django_connections[alias]
    params = dj_connection.get_connection_params()
    return await asyncpg.create_pool(**params)


async def get_pool(alias: Optional[ConnectionAlias] = None) -> asyncpg.Pool:
    """
    Получение пула подключений asyncpg.Pool, аналогичного подключению
    alias в django

    Args:
        alias: псевдоним БД в django,
            если не передан - используется БД по-умолчанию

    Returns:
        пул подключений asyncpg.Pool

    """
    if alias is None:
        alias = DEFAULT_DB_ALIAS

    if alias not in connections:
        connections[alias] = await _make_connection(alias)

    connection = connections[alias]

    return connection


async def get_connection(alias: Optional[ConnectionAlias] = None) -> asyncpg.Connection:
    """
    Получение подключения asyncpg.Connection, аналогичного подключению
    alias в django

    Args:
        alias: псевдоним БД в django,
            если не передан - используется БД по-умолчанию

    Returns:
        подключение asyncpg.Connection

    """
    pool = await get_pool(alias)
    return pool.acquire()


async def truncate_table(conn: ConnectionInterface, model: Model):
    """
    Сброс данных из таблицы модели model

    Args:
        conn: asyncpg подключение к БД (asyncpg.Connection, либо asyncpg.Pool)
        model: django-модель

    """
    db_table = model._meta.db_table
    query = f'TRUNCATE TABLE {db_table} RESTART IDENTITY CASCADE'
    await conn.execute(query)


def get_unique_violation_details(
    exc: asyncpg.exceptions.UniqueViolationError,
) -> tuple[Model, Field, Any]:
    """
    Получение подробностей исключения UniqueViolationError

    Args:
        exc: исключение UniqueViolationError

    Returns:
        кортеж из django-модели, поля и значения, из-за которых произошла ошибка

    """

    # re.search без re.compile(pattern)
    # в сравнении с re.fullmatch по полному тексту сообшения и/или
    # с использованием прекомпилированного за пределами функции паттерна
    # работает быстрее всего в данном случае
    match = re.search(r'\((?P<column>.+)\)=\((?P<value>.+)\)', exc.detail)

    column = match.group('column')
    value = match.group('value')

    # value нужно привести к типу столбца
    model = next(
        model for model
        in apps.get_models()
        if model._meta.db_table == exc.table_name
    )
    field = next(
        field for field
        in model._meta.get_fields()
        if field.get_attname_column()[1] == column
    )
    value = field.to_python(value)

    return model, field, value



from django.contrib.postgres.indexes import (
    HashIndex,
)

from m3_gar.base_models import (
    Houses as BaseHouses,
    HouseTypes as BaseHouseTypes,
)
from m3_gar.models.op_type import (
    OperationTypes,
)
from m3_gar.models.reestr import (
    ReestrObjects,
)
from m3_gar.models.util import (
    RegionCodeModelMixin,
    add_params,
    make_fk,
)


__all__ = ['Houses', 'HouseTypes', 'AddhouseTypes']


class HouseTypes(BaseHouseTypes):
    """
    Сведения по типам домов
    """
    class Meta:
        verbose_name = 'Тип дома'
        verbose_name_plural = 'Типы домов'


class AddhouseTypes(BaseHouseTypes):
    """
    Сведения по дополнительным типам домов
    """
    class Meta:
        verbose_name = 'Тип дома'
        verbose_name_plural = 'Типы домов'


class Houses(BaseHouses, RegionCodeModelMixin):
    """
    Сведения по номерам домов улиц городов и населенных пунктов
    """

    level = 10

    class Meta:
        verbose_name = 'Номер дома'
        verbose_name_plural = 'Номера домов'

        indexes = (
            HashIndex(
                fields=('objectguid',),
            ),
        )

make_fk(Houses, 'housetype', to=HouseTypes, null=True, blank=True)
make_fk(Houses, 'addtype1', to=AddhouseTypes, null=True, blank=True)
make_fk(Houses, 'addtype2', to=AddhouseTypes, null=True, blank=True)
make_fk(Houses, 'opertypeid', to=OperationTypes)
make_fk(Houses, 'objectid', to=ReestrObjects)

add_params(Houses, 'm3_gar.HousesParams')



# Generated by Django 2.2.28 on 2022-04-15 14:52

import django.contrib.postgres.indexes
from django.contrib.postgres.operations import (
    BtreeGinExtension,
    TrigramExtension,
)
from django.db import (
    migrations,
    models,
)


class Migration(migrations.Migration):

    dependencies = [
        ('m3_gar', '0011_auto_20220323_1022'),
    ]

    operations = [
        TrigramExtension(),
        BtreeGinExtension(),
        migrations.AddField(
            model_name='admhierarchy',
            name='name_with_parents',
            field=models.CharField(blank=True, max_length=500, null=True, verbose_name='Полное наименование'),
        ),
        migrations.AddField(
            model_name='munhierarchy',
            name='name_with_parents',
            field=models.CharField(blank=True, max_length=500, null=True, verbose_name='Полное наименование'),
        ),
        migrations.AddIndex(
            model_name='admhierarchy',
            index=django.contrib.postgres.indexes.GinIndex(fields=['name_with_parents'], name='m3_gar_admh_name_wi_5f5f5a_gin'),
        ),
        migrations.AddIndex(
            model_name='munhierarchy',
            index=django.contrib.postgres.indexes.GinIndex(fields=['name_with_parents'], name='m3_gar_munh_name_wi_380974_gin'),
        ),
    ]



import codecs
from pathlib import (
    Path,
)

from setuptools import (
    find_packages,
)

from m3_gar.version import (
    __version__,
)


def read(fn):
    return codecs.open(Path(__file__).resolve().parent / fn).read()


def make_setup_kwargs():
    return dict(
        name='m3-gar',
        version=__version__,
        author='BARS Group',
        author_email='bars@bars.group',
        description='GAR Django integration for m3',
        long_description=read('README.rst'),
        license='MIT license',
        install_requires=[
            'Django>=3.2',
            'requests',
            'lxml',
            'progress',
            'gitpython',
            'twine',
            'asyncpg',
            'uvloop',
            'django-cte==1.1.5',
            'm3-gar-constants>=1.0.4',
        ],
        packages=find_packages(exclude=['tests', 'test_project']),
        include_package_data=True,
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: MIT License',
            'Natural Language :: Russian',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.9',
            'Topic :: Software Development :: Libraries :: Python Modules',
        ],
    )



# Generated by Django 3.2.25 on 2024-03-29 15:39

import django.contrib.postgres.indexes
from django.db import (
    migrations,
)


class Migration(migrations.Migration):

    dependencies = [
        ('m3_gar', '0016_alter_normativedocs_orgname'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='apartments',
            index=django.contrib.postgres.indexes.HashIndex(fields=['objectguid'], name='m3_gar_apar_objectg_0e459a_hash'),
        ),
    ]



from django.apps import (
    apps,
)
from django.db import (
    models,
)

import django_cte


def make_fk(model, field_name, **kwargs):
    field = model._meta.get_field(field_name)
    _, db_column = field.get_attname_column()

    unique = kwargs.pop('unique', False)
    kwargs.setdefault('on_delete', models.DO_NOTHING)
    kwargs.setdefault('related_name', '+')
    kwargs.setdefault('verbose_name', field.verbose_name)
    kwargs.setdefault('null', field.null)
    kwargs.setdefault('blank', field.blank)
    kwargs.setdefault('db_column', db_column)

    if unique:
        new_field_type = models.OneToOneField
    else:
        new_field_type = models.ForeignKey

    new_field = new_field_type(**kwargs)

    model._meta.local_fields.remove(field)
    new_field.contribute_to_class(model, field_name)


def add_params(addr_model, params_model):
    @property
    def params(self):
        Param = apps.get_model(params_model)

        class ParamManager(Param._default_manager.__class__):
            def get_queryset(manager):
                return super().get_queryset().filter(objectid=self.objectid_id)

        manager = ParamManager()
        manager.model = Param

        return manager

    addr_model.params = params


class RegionCodeModelMixin(models.Model):
    region_code = models.SmallIntegerField(verbose_name='Код региона')

    class Meta:
        abstract = True


class HierarchyMixin(models.Model):
    """
    Класс предоставляет часть интерфейса иерархической модели
    """
    parent_link_field = 'parentobjid'
    objects = django_cte.CTEManager()

    class Meta:
        abstract = True

    def get_ancestors(self, include_self=False):
        """
        Получение всех записей, стоящих выше данной в иерархии
        по parent_id (предков).
        """
        cls = self.__class__

        def make_cte(cte):
            col_parent_id = getattr(cte.col, cls.parent_link_field)

            query = cls.objects.filter(
                id=self.id
            ).values(
                "id",
                cls.parent_link_field,
                level=models.Value(0, output_field=models.IntegerField()),
            ).union(
                cte.join(
                    cls.objects.filter(isactive=True),
                    objectid=col_parent_id,
                ).values(
                    "id",
                    cls.parent_link_field,
                    level=(
                            models.ExpressionWrapper(cte.col.level, output_field=models.IntegerField()) -
                            models.Value(1, output_field=models.IntegerField())
                    )
                ),
                all=True,
            )

            return query

        ctexpr = django_cte.With.recursive(make_cte)

        ancestors = ctexpr.join(
            cls.objects.all(), id=ctexpr.col.id
        ).with_cte(
            ctexpr
        ).annotate(
            level=ctexpr.col.level,
        ).order_by("level")

        if not include_self:
            ancestors = ancestors.exclude(id=self.id)

        return ancestors

    @property
    def has_children(self):
        return self.__class__.objects.filter(
            isactive=True,
            objectid__levelid=self.objectid.levelid_id + 1,
            **{self.parent_link_field: self.objectid_id},
        ).exists()



from typing import (
    TYPE_CHECKING,
)

from m3_gar.enums import (
    VersionUpdateStatus,
)
from m3_gar.importer.client import (
    client,
)
from m3_gar.importer.exceptions import (
    VersionAlreadyProcessedError,
)
from m3_gar.importer.signals import (
    post_fetch_version,
    pre_fetch_version,
)
from m3_gar.models import (
    Version,
)


if TYPE_CHECKING:
    from m3_gar.importer.client import (
        DownloadFileInfo,
    )


def get_or_create_version(file_info: 'DownloadFileInfo') -> tuple[Version, bool]:
    """Создаёт и/или возвращает запись Version по информации об обновлении."""
    version, created = Version.objects.get_or_create(
        ver=file_info.version_id,
        defaults={
            'dumpdate': file_info.date,
            'complete_xml_url': file_info.gar_xml_full_url,
            'delta_xml_url': file_info.gar_xml_delta_url,
        }
    )

    return version, created


def check_version_is_processed(version: Version) -> None:
    """Проверка версии на признак обработки.

    Raises:
        VersionAlreadyProcessedError: если версия уже была обработана ранее
    """
    if version.status in VersionUpdateStatus.get_processed_statuses():
        raise VersionAlreadyProcessedError(f'Версия {version} уже была обработана ранее!')


def update_version_status(version: Version, status: VersionUpdateStatus) -> None:
    """Изменение статуса у версии обновления."""
    check_version_is_processed(version)

    version.status = status
    version.save()


def fetch_version_info(update_all: bool = False) -> None:
    """Получение информации о версии файлов обновления.

    Получает информацию о файлах с обновлениями с сайта ФИАС и создаёт по ним
    соответствующие записи модели Version.

    Args:
        update_all: Обновляет информацию по всем существующим записям Version
    """
    pre_fetch_version.send(object.__class__)

    all_download_file_info = client.get_all_download_file_info()

    for file_info in all_download_file_info:
        version, created = get_or_create_version(file_info)

        if update_all and not created:
            version.complete_xml_url = file_info.gar_xml_full_url
            version.delta_xml_url = file_info.gar_xml_delta_url

            if version.dumpdate < file_info.date:
                version.dumpdate = file_info.date

            version.save()

    post_fetch_version.send(object.__class__)



from typing import (
    TYPE_CHECKING,
)

from m3_gar.importer.signals import (
    post_load,
    pre_load,
)
from m3_gar.importer.source.wrapper import (
    SourceWrapper,
)
from m3_gar.importer.table import (
    TableFactory,
)
from m3_gar.models import (
    Version,
)


if TYPE_CHECKING:
    from m3_gar.importer.table import (
        XMLTable,
    )


class TableList:
    """
    Базовый класс для работы с набором импортируемых файлов
    """

    wrapper_class = SourceWrapper
    wrapper = None

    table_list = None
    date = None
    version_info = None

    def __init__(self, src, version=None, tempdir=None):
        self.info_version = version
        self.src = src
        self.tempdir = tempdir

        if version is not None:
            assert isinstance(version, Version), 'version must be an instance of Version model'

            self.date = version.dumpdate

        pre_load.send(sender=self.__class__, src=src)
        self.wrapper = self.load_data(src)
        post_load.send(sender=self.__class__, wrapper=self.wrapper)

    def load_data(self, source):
        return self.wrapper_class(source=source)

    def get_table_list(self):
        return self.wrapper.get_file_list()

    @property
    def tables(self) -> dict[str, list['XMLTable']]:
        if self.table_list is None:
            self.table_list = {}
            for filename in self.get_table_list():
                table = TableFactory.parse(filename=filename)
                if table is None:
                    continue
                self.table_list.setdefault(table.name, []).append(table)

        return self.table_list

    def get_date_info(self, name):
        return self.wrapper.get_date_info(filename=name)

    @property
    def dump_date(self):
        if self.date is None:
            first_name = self.get_table_list()[0]
            self.date = self.get_date_info(first_name)

        return self.date

    def open(self, filename):
        return self.wrapper.open(filename=filename)

    @property
    def version(self):
        if self.version_info is None:
            self.version_info = Version.objects.nearest_by_date(self.dump_date)
        return self.version_info



# Generated by Django 3.2.25 on 2024-04-11 11:00

from django.db import (
    migrations,
)


class Migration(migrations.Migration):

    dependencies = [
        ('m3_gar', '0020_update_version_status'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='version',
            name='processed',
        ),
    ]



# Generated by Django 3.2.4 on 2022-03-23 10:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('m3_gar', '0010_auto_20220225_0934'),
    ]

    operations = [
        migrations.AlterField(
            model_name='admhierarchy',
            name='path',
            field=models.TextField(blank=True, null=True, verbose_name='Материализованный путь к объекту (полная иерархия)'),
        ),
        migrations.AlterField(
            model_name='munhierarchy',
            name='path',
            field=models.TextField(blank=True, null=True, verbose_name='Материализованный путь к объекту (полная иерархия)'),
        ),
    ]



# Generated by Django 3.2.4 on 2021-09-06 07:44

import django.db.models.deletion
from django.db import (
    migrations,
    models,
)


class Migration(migrations.Migration):

    dependencies = [
        ('m3_gar', '0002_auto_20210707_1352'),
    ]

    operations = [
        migrations.CreateModel(
            name='AddhouseTypes',
            fields=[
                ('id', models.IntegerField(primary_key=True, serialize=False, verbose_name='Идентификатор')),
                ('name', models.CharField(max_length=50, verbose_name='Наименование')),
                ('shortname', models.CharField(blank=True, max_length=50, null=True, verbose_name='Краткое наименование')),
                ('desc', models.CharField(blank=True, max_length=250, null=True, verbose_name='Описание')),
                ('updatedate', models.DateField(verbose_name='Дата внесения (обновления) записи')),
                ('startdate', models.DateField(verbose_name='Начало действия записи')),
                ('enddate', models.DateField(verbose_name='Окончание действия записи')),
                ('isactive', models.BooleanField(verbose_name='Статус активности')),
            ],
            options={
                'verbose_name': 'Тип дома',
                'verbose_name_plural': 'Типы домов',
            },
        ),
        migrations.AddField(
            model_name='addrobj',
            name='region_code',
            field=models.SmallIntegerField(default=0, verbose_name='Код региона'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='addrobjdivision',
            name='region_code',
            field=models.SmallIntegerField(default=0, verbose_name='Код региона'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='admhierarchy',
            name='region_code',
            field=models.SmallIntegerField(default=0, verbose_name='Код региона'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='apartments',
            name='region_code',
            field=models.SmallIntegerField(default=0, verbose_name='Код региона'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='carplaces',
            name='region_code',
            field=models.SmallIntegerField(default=0, verbose_name='Код региона'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='changehistory',
            name='region_code',
            field=models.SmallIntegerField(default=0, verbose_name='Код региона'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='houses',
            name='region_code',
            field=models.SmallIntegerField(default=0, verbose_name='Код региона'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='munhierarchy',
            name='region_code',
            field=models.SmallIntegerField(default=0, verbose_name='Код региона'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='normativedocs',
            name='region_code',
            field=models.SmallIntegerField(default=0, verbose_name='Код региона'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='param',
            name='region_code',
            field=models.SmallIntegerField(default=0, verbose_name='Код региона'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='reestrobjects',
            name='region_code',
            field=models.SmallIntegerField(default=0, verbose_name='Код региона'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='rooms',
            name='region_code',
            field=models.SmallIntegerField(default=0, verbose_name='Код региона'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='steads',
            name='region_code',
            field=models.SmallIntegerField(default=0, verbose_name='Код региона'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='addrobj',
            name='objectid',
            field=models.ForeignKey(db_column='objectid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.reestrobjects', verbose_name='Глобальный уникальный идентификатор адресного объекта типа INTEGER'),
        ),
        migrations.AlterField(
            model_name='addrobj',
            name='opertypeid',
            field=models.ForeignKey(db_column='opertypeid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.operationtypes', verbose_name='Статус действия над записью – причина появления записи'),
        ),
        migrations.AlterField(
            model_name='addrobjdivision',
            name='childid',
            field=models.ForeignKey(db_column='childid', db_constraint=False, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.reestrobjects', verbose_name='Дочерний ID'),
        ),
        migrations.AlterField(
            model_name='addrobjdivision',
            name='parentid',
            field=models.ForeignKey(db_column='parentid', db_constraint=False, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.reestrobjects', verbose_name='Родительский ID'),
        ),
        migrations.AlterField(
            model_name='admhierarchy',
            name='objectid',
            field=models.ForeignKey(db_column='objectid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.reestrobjects', verbose_name='Глобальный уникальный идентификатор объекта'),
        ),
        migrations.AlterField(
            model_name='admhierarchy',
            name='parentobjid',
            field=models.ForeignKey(blank=True, db_column='parentobjid', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.reestrobjects', verbose_name='Идентификатор родительского объекта'),
        ),
        migrations.AlterField(
            model_name='apartments',
            name='aparttype',
            field=models.ForeignKey(blank=True, db_column='aparttype', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.apartmenttypes', verbose_name='Тип комнаты'),
        ),
        migrations.AlterField(
            model_name='apartments',
            name='objectid',
            field=models.ForeignKey(db_column='objectid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.reestrobjects', verbose_name='Глобальный уникальный идентификатор объекта типа INTEGER'),
        ),
        migrations.AlterField(
            model_name='apartments',
            name='opertypeid',
            field=models.ForeignKey(db_column='opertypeid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.operationtypes', verbose_name='Статус действия над записью – причина появления записи'),
        ),
        migrations.AlterField(
            model_name='carplaces',
            name='objectid',
            field=models.ForeignKey(db_column='objectid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.reestrobjects', verbose_name='Глобальный уникальный идентификатор объекта типа INTEGER'),
        ),
        migrations.AlterField(
            model_name='carplaces',
            name='opertypeid',
            field=models.ForeignKey(db_column='opertypeid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.operationtypes', verbose_name='Статус действия над записью – причина появления записи'),
        ),
        migrations.AlterField(
            model_name='changehistory',
            name='opertypeid',
            field=models.ForeignKey(db_column='opertypeid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.operationtypes', verbose_name='Тип операции'),
        ),
        migrations.AlterField(
            model_name='houses',
            name='housetype',
            field=models.ForeignKey(blank=True, db_column='housetype', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.housetypes', verbose_name='Основной тип дома'),
        ),
        migrations.AlterField(
            model_name='houses',
            name='objectid',
            field=models.ForeignKey(db_column='objectid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.reestrobjects', verbose_name='Глобальный уникальный идентификатор объекта типа INTEGER'),
        ),
        migrations.AlterField(
            model_name='houses',
            name='opertypeid',
            field=models.ForeignKey(db_column='opertypeid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.operationtypes', verbose_name='Статус действия над записью – причина появления записи'),
        ),
        migrations.AlterField(
            model_name='munhierarchy',
            name='objectid',
            field=models.ForeignKey(db_column='objectid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.reestrobjects', verbose_name='Глобальный уникальный идентификатор адресного объекта '),
        ),
        migrations.AlterField(
            model_name='munhierarchy',
            name='parentobjid',
            field=models.ForeignKey(blank=True, db_column='parentobjid', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.reestrobjects', verbose_name='Идентификатор родительского объекта'),
        ),
        migrations.AlterField(
            model_name='normativedocs',
            name='kind',
            field=models.ForeignKey(db_column='kind', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.normativedocskinds', verbose_name='Вид документа'),
        ),
        migrations.AlterField(
            model_name='normativedocs',
            name='type',
            field=models.ForeignKey(db_column='type', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.normativedocstypes', verbose_name='Тип документа'),
        ),
        migrations.AlterField(
            model_name='param',
            name='typeid',
            field=models.ForeignKey(db_column='typeid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.paramtypes', verbose_name='Тип параметра'),
        ),
        migrations.AlterField(
            model_name='reestrobjects',
            name='levelid',
            field=models.ForeignKey(db_column='levelid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.objectlevels', verbose_name='Уровень объекта'),
        ),
        migrations.AlterField(
            model_name='rooms',
            name='objectid',
            field=models.ForeignKey(db_column='objectid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.reestrobjects', verbose_name='Глобальный уникальный идентификатор объекта типа INTEGER'),
        ),
        migrations.AlterField(
            model_name='rooms',
            name='opertypeid',
            field=models.ForeignKey(db_column='opertypeid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.operationtypes', verbose_name='Статус действия над записью – причина появления записи'),
        ),
        migrations.AlterField(
            model_name='rooms',
            name='roomtype',
            field=models.ForeignKey(blank=True, db_column='roomtype', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.roomtypes', verbose_name='Тип комнаты или офиса'),
        ),
        migrations.AlterField(
            model_name='steads',
            name='objectid',
            field=models.ForeignKey(db_column='objectid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.reestrobjects', verbose_name='Глобальный уникальный идентификатор объекта типа INTEGER'),
        ),
        migrations.AlterField(
            model_name='steads',
            name='opertypeid',
            field=models.ForeignKey(db_column='opertypeid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.operationtypes', verbose_name='Статус действия над записью – причина появления записи'),
        ),
        migrations.AlterField(
            model_name='houses',
            name='addtype1',
            field=models.ForeignKey(blank=True, db_column='addtype1', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.addhousetypes', verbose_name='Дополнительный тип дома 1'),
        ),
        migrations.AlterField(
            model_name='houses',
            name='addtype2',
            field=models.ForeignKey(blank=True, db_column='addtype2', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.addhousetypes', verbose_name='Дополнительный тип дома 2'),
        ),
    ]



# Generated by Django 3.2.25 on 2024-03-29 19:29

from django.db import (
    migrations,
)

import m3_gar.models.indexes


class Migration(migrations.Migration):

    dependencies = [
        ('m3_gar', '0017_apartments_m3_gar_apar_objectg_0e459a_hash'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='addrobj',
            index=m3_gar.models.indexes.UpperGinIndex(fastupdate=True, fields=['name_with_typename'], name='addrobj_name_with_typename_gin', opclasses=['gin_trgm_ops']),
        ),
    ]



class ImporterError(Exception):
    pass


class DatabaseNotEmptyError(ImporterError):
    pass


class DownloadFileInfoError(ImporterError):
    """Ошибка связанная с обработкой информации о файлах обновлений с сайта ФИАС."""
    pass


class UpdateVersionError(ImporterError):
    """Общая ошибка связанная с обработкой записей обновлений."""
    pass


class VersionAlreadyProcessedError(UpdateVersionError):
    """Ошибка при попытке обработать версию, которая уже была обработана ранее."""
    pass


class VersionDoesNotHaveUrl(UpdateVersionError):
    """У версии обновления отсутствуют ссылки на файл с данными."""
    pass



# Generated by Django 3.2.25 on 2024-04-10 08:50

from django.db import (
    migrations,
    models,
)


class Migration(migrations.Migration):

    dependencies = [
        ('m3_gar', '0018_addrobj_addrobj_name_with_typename_gin'),
    ]

    operations = [
        migrations.AddField(
            model_name='version',
            name='status',
            field=models.CharField(choices=[('new', 'Новое'), ('skipped', 'Пропущено'), ('error', 'Ошибка'), ('finished', 'Завершено')], default='new', max_length=20, verbose_name='Статус обновления'),
        ),
    ]



from django.dispatch import (
    Signal,
)


pre_fetch_version = Signal(providing_args=[])
post_fetch_version = Signal(providing_args=[])

pre_download = Signal(providing_args=['url'])
post_download = Signal(providing_args=['url', 'path'])

pre_unpack = Signal(providing_args=['archive'])
post_unpack = Signal(providing_args=['archive', 'dst'])

pre_load = Signal(providing_args=['src'])
post_load = Signal(providing_args=['wrapper'])

pre_drop_indexes = Signal(providing_args=['table'])
post_drop_indexes = Signal(providing_args=['table'])

pre_restore_indexes = Signal(providing_args=['table'])
post_restore_indexes = Signal(providing_args=['table'])

pre_import_table = Signal(providing_args=['table'])
post_import_table = Signal(providing_args=['table'])

pre_import = Signal(providing_args=['msg'])
post_import = Signal(providing_args=['msg'])

pre_update = Signal(providing_args=['before', 'after'])
post_update = Signal(providing_args=['before', 'after'])



from copy import (
    deepcopy,
)

from django.db import (
    connections,
    models,
)

from m3_gar import (
    config,
)


class ConstraintsEditor:

    def __init__(self, *, enable, fk, unique, index):
        self.enable = enable
        self.fk = fk
        self.unique = unique
        self.index = index

    def get_simple_field(self, field):
        """
        Возвращает упрощённую копию переданного поля с отключенными индексами,
        уникальными ограничениями и проверками внешних ключей
        """

        new_field = deepcopy(field)

        if self.unique:
            new_field._unique = False

        if self.index:
            new_field.db_index = False

        if self.fk and any((
            isinstance(field, models.ForeignKey),
            isinstance(field, models.ManyToManyField) and field.remote_field.through is None,
        )):
            new_field.db_constraint = False

        return new_field

    def get_simple_fields(self, model):
        """
        Возвращает пары (поле, упрощённая копия поля) для всех собственных полей модели
        """

        fields = model._meta.get_fields(include_parents=False, include_hidden=False)

        for field in fields:
            if field.concrete:
                yield field, self.get_simple_field(field)

    @staticmethod
    def alter_field(model, field_from, field_to):
        """
        Производит преобразование поля field_from на field_to модели model
        в БД для записей ГАР

        Args:
            model: модель
            field_from: исходное поле
            field_to: новое поле

        Returns:
            список выполненных sql-запросов

        """
        con = connections[config.DATABASE_ALIAS]
        ed = con.schema_editor(collect_sql=True)

        ed.alter_field(model, field_from, field_to)

        return ed.collected_sql

    def change_model_constraints(self, model):
        """
        Удаляет ограничения и индексы модели
        """
        for field, simple_field in self.get_simple_fields(model=model):

            field_from, field_to = field, simple_field

            if self.enable:
                field_from, field_to = field_to, field_from

            yield self.alter_field(model=model, field_from=field_from, field_to=field_to)



# Generated by Django 3.2.25 on 2024-04-10 08:57

from django.db import migrations

from m3_gar.enums import (
    VersionUpdateStatus,
)


def update_version_status(apps, schema_editor):
    """Обновление значений поля Version.status."""
    Version = apps.get_model('m3_gar', 'Version')

    Version.objects.filter(
        processed=True,
    ).exclude(
        delta_xml_url='',
        complete_xml_url='',
    ).update(
        status=VersionUpdateStatus.FINISHED,
    )

    Version.objects.filter(
        processed=True,
        delta_xml_url='',
        complete_xml_url='',
    ).update(
        status=VersionUpdateStatus.SKIPPED,
    )


def set_processed_by_status(apps, schema_editor):
    """Обновляет поле processed по статусу."""
    Version = apps.get_model('m3_gar', 'Version')

    Version.objects.filter(
        status__in=VersionUpdateStatus.get_processed_statuses(),
    ).update(processed=True)


class Migration(migrations.Migration):

    dependencies = [
        ('m3_gar', '0019_version_status'),
    ]

    operations = [
        migrations.RunPython(update_version_status, set_processed_by_status)
    ]



import asyncio
import os
import queue
from concurrent.futures import (
    ThreadPoolExecutor,
)
from contextlib import (
    asynccontextmanager,
)
from datetime import (
    timedelta,
)
from operator import (
    attrgetter,
)
from typing import (
    TYPE_CHECKING,
    Optional,
)

import asyncpg
import uvloop
from asgiref.sync import (
    async_to_sync,
)
from django.core.exceptions import (
    ValidationError,
)
from django.core.validators import (
    URLValidator,
)
from django.db import (
    models,
    transaction,
)
from django.utils import (
    timezone,
)

from m3_gar import (
    config,
)
from m3_gar.compat import (
    nullcontext,
)
from m3_gar.consts import (
    VERSION_WITHOUT_URL_TIMEOUT_HOURS,
)
from m3_gar.enums import (
    VersionUpdateStatus,
)
from m3_gar.importer import (
    db_wrapper,
)
from m3_gar.importer.consts import (
    DEFAULT_BULK_LIMIT,
)
from m3_gar.importer.exceptions import (
    DatabaseNotEmptyError,
    ImporterError,
    VersionDoesNotHaveUrl,
)
from m3_gar.importer.loader import (
    TableUpdater,
)
from m3_gar.importer.signals import (
    post_import,
    post_update,
    pre_import,
    pre_update,
)
from m3_gar.importer.source import *
from m3_gar.importer.source.exceptions import (
    TableListLoadingError,
)
from m3_gar.importer.version import (
    check_version_is_processed,
    update_version_status,
)
from m3_gar.models import (
    Version,
)
from m3_gar.util import (
    get_table_names_from_models,
)


if TYPE_CHECKING:
    from m3_gar.importer.source.tablelist import (
        TableList,
    )


def get_tablelist(
    path: str,
    version: Optional[Version] = None,
    tempdir: Optional[str] = None,
) -> 'TableList':
    """Возвращает объект одной из реализаций TableList в зависимости от параметров."""
    if os.path.isfile(path):
        # Путь до файла
        tablelist = LocalArchiveTableList(src=path, version=version, tempdir=tempdir)

    elif os.path.isdir(path):
        # Путь до директории
        tablelist = DirectoryTableList(src=path, version=version, tempdir=tempdir)

    else:
        # Если не файл и не директория, то полагаем, что передан URL
        try:
            URLValidator()(path)
        except ValidationError:
            tablelist = None
        else:
            tablelist = RemoteArchiveTableList(src=path, version=version, tempdir=tempdir)

    if not tablelist:
        raise TableListLoadingError(f'Путь `{path}` не является валидным источником для обновления')

    return tablelist


def get_table_names(tables):
    return tables if tables else get_table_names_from_models()


def load_complete_data(**kwargs):
    no_transaction = kwargs.get('no_transaction', False)

    if no_transaction:
        atomic = nullcontext()
    else:
        atomic = transaction.atomic(using=config.DATABASE_ALIAS)

    with atomic:
        _load_complete_data(**kwargs)


def _load_complete_data(
    *,
    path=None,
    truncate=None,
    no_transaction=False,
    limit=DEFAULT_BULK_LIMIT,
    tables=None,
    tempdir=None,
):
    """
    Загрузка полного архива БД ГАР

    Args:
        path: путь на диске или URL до загруженного архива или директории с
            распакованным архивом. Если передан None, используется известный
            URL до новейшей версии архива.
        truncate: очищать ли уже существующие в БД данные.
            При truncate=None и наличии данных в БД, выбрасывается
            исключение DatabaseNotEmptyError.
            При truncate=True, данные в БД будут предварительно очищены.
            При truncate=False, наличие существующих данных в БД игнорируется.
        no_transaction: отключение заворачивания всего процесса загрузки в транзакцию
        limit: количество записей для пакетного сохранения.
        tables: список импортируемых таблиц. Если передан None, импортируются
            все данные
        tempdir: путь до временной директории для загрузки и распаковки архивов

    """

    tablelist = get_tablelist(
        path=path,
        tempdir=tempdir,
    )

    pre_import.send(
        sender=object.__class__,
        version=tablelist.version,
    )

    table_names = get_table_names(tables)
    table_names = [
        # Пропускаем таблицы, которых нет в архиве
        tbl for tbl in table_names
        if tbl in tablelist.tables
    ]

    if truncate is True:
        Version.objects.update(status=VersionUpdateStatus.NEW)

    if truncate is not False and Version.objects.filter(processed=True).exists():
        raise DatabaseNotEmptyError()

    @async_to_sync
    async def do_the_async_load():

        @asynccontextmanager
        async def make_conn():
            if no_transaction:
                # Если работаем без транзакции, можно делать запросы прямо к
                # пулу соединений, Lock для работы с ним не нужен
                conn = await db_wrapper.get_pool(config.DATABASE_ALIAS)
                conn_lock = nullcontext()

                async with conn:
                    yield conn, conn_lock

            else:
                # Если работаем в транзакции, обязательно захватываем
                # соединение и выдаём Lock на работу с ним
                conn = await db_wrapper.get_connection(config.DATABASE_ALIAS)
                conn_lock = asyncio.Lock()

                async with conn:
                    async with conn.transaction():
                        yield conn, conn_lock

        async with make_conn() as (conn, conn_lock):

            for tbl in table_names:
                table = tablelist.tables[tbl][0]

                if truncate is True:
                    await db_wrapper.truncate_table(conn, table.model)

            read_pool = ThreadPoolExecutor()
            read_queue = queue.Queue(maxsize=limit)
            objects_lists = {}
            objects_counters = {}
            processing_tables = {}
            create_tasks = []
            create_failure: Optional[asyncio.Task] = None

            for tbl in table_names:
                for table in tablelist.tables[tbl]:
                    objects_lists.setdefault(tbl, [])
                    objects_counters.setdefault(tbl, 0)
                    processing_tables.setdefault(tbl, set()).add(table.filename)

                    read_pool.submit(
                        _load_rows_into_queue,
                        tablelist=tablelist,
                        table=table,
                        results=read_queue,
                    )

            async def create(objects):
                nonlocal create_failure

                try:
                    return await _create(objects)

                except Exception:
                    create_failure = asyncio.current_task()
                    raise

                finally:
                    del objects


            async def _create(objects):
                if objects:
                    model = objects[0]._meta.model

                    attnames_columns = [
                        field.get_attname_column()
                        for field in model._meta.concrete_fields
                    ]
                    attnames = [attname for attname, column in attnames_columns]
                    columns = [column for attname, column in attnames_columns]

                    model_attrgetter = attrgetter(*attnames)

                    try:
                        async with conn_lock:
                            await conn.copy_records_to_table(
                                table_name=model._meta.db_table,
                                columns=columns,
                                records=(
                                    model_attrgetter(obj) for obj in objects
                                ),
                            )

                    except asyncpg.exceptions.UniqueViolationError as e:
                        model, field, value = db_wrapper.get_unique_violation_details(e)
                        attname = field.get_attname()
                        new_objects = [
                            object for object in objects
                            if getattr(object, attname) != value
                        ]

                        if len(new_objects) == len(objects):
                            raise

                        print(
                            f'Found duplicate {model.__name__} record {attname}={value}. '
                            f'Retrying skipping this record.'
                        )

                        create_tasks.append(asyncio.create_task(
                            create(new_objects)
                        ))

            while True:
                await asyncio.sleep(0)

                if create_failure:
                    await create_failure

                try:
                    row, table = read_queue.get(block=False)
                except queue.Empty:
                    continue

                if isinstance(row, models.Model):
                    objects_lists[table.name].append(row)
                    objects_counters[table.name] += 1
                    if objects_counters[table.name] % limit == 0:
                        print(f'T: {table.name}\tL: {objects_counters[table.name]}')

                        create_tasks.append(asyncio.create_task(
                            create(objects_lists[table.name])
                        ))
                        objects_lists[table.name] = []

                elif row is StopIteration:
                    processing_tables[table.name].remove(table.filename)
                    print(f'F: {table.filename}\tDONE')

                    if not processing_tables[table.name]:
                        print(f'T: {table.name}\tL: {objects_counters[table.name]}')
                        print(f'T: {table.name}\tDONE')

                        create_tasks.append(asyncio.create_task(
                            create(objects_lists[table.name])
                        ))
                        del processing_tables[table.name]

                    if not processing_tables:
                        print(f'ALL DONE')
                        break

                elif isinstance(row, Exception):
                    raise row

                else:
                    raise RuntimeError

            print('Awaiting pending database write tasks')
            for i, t in enumerate(create_tasks, start=1):
                if not t.done():
                    await t
                    print(f'{i} out of {len(create_tasks)} tasks done')

    uvloop.install()
    do_the_async_load()

    # только если path был равен None - мы имеем право зарегистрировать все версии как обработанные, потому что в этом
    # случае автоматически разворачивается дамп с последней версии. иначе мы этого точно знать не можем, но попробуем
    # определить версию из url (если он был передан)
    if path:
        error_msg = (
            'Unknown uploaded version. Please set attribute status=VersionUpdateStatus.FINISHED to all instances of '
            'Version model that less or equal to uploaded version yourself, or next update will be more time consuming'
        )
        try:
            URLValidator()(path)
        except ValidationError:
            print(error_msg)
        else:
            try:
                url_ver = int(path.split('/')[-2].replace('.', ''))

                Version.objects.filter(
                    ver__lte=url_ver,
                ).exclude(
                    status=VersionUpdateStatus.SKIPPED,
                ).update(
                    status=VersionUpdateStatus.FINISHED,
                )
            except (IndexError, ValueError):
                print(error_msg)

    else:
        Version.objects.exclude(
            status=VersionUpdateStatus.SKIPPED,
        ).update(
            status=VersionUpdateStatus.FINISHED,
        )

    post_import.send(
        sender=object.__class__,
        version=tablelist.version,
    )


def _load_rows_into_queue(tablelist, table, results):
    def put_result(row):
        results.put((row, table))

    try:
        for row in table.rows(tablelist=tablelist):
            put_result(row)
    except Exception as e:
        put_result(e)

    put_result(StopIteration)


@transaction.atomic(using=config.DATABASE_ALIAS)
def update_data(
    version: Version,
    limit: int = 1000,
    tables: Optional[tuple[str, ...]] = None,
    tempdir: Optional[str] = None,
) -> None:
    """
    Загрузка дельта-архива БД ГАР.

    Args:
        version: объект версии архива
        limit: количество записей для пакетного сохранения. Не влияет на
            обновляемые объекты.
        tables: список импортируемых таблиц. Если передан None, импортируются
            все данные
        tempdir: путь до временной директории для загрузки и распаковки архивов
    """
    check_version_is_processed(version)

    tablelist = get_tablelist(path=version.delta_xml_url, version=version, tempdir=tempdir)

    for table_name in get_table_names(tables):
        # Пропускаем таблицы, которых нет в архиве
        if table_name not in tablelist.tables:
            continue

        for table in tablelist.tables[table_name]:
            loader = TableUpdater(limit=limit)
            loader.load(tablelist=tablelist, table=table)

    update_version_status(version, VersionUpdateStatus.FINISHED)

    # удаляем временный файл с архивом обновления
    temp_file_name = tablelist.wrapper.source.filename
    if os.path.exists(temp_file_name):
        os.remove(temp_file_name)


def auto_update_data(
    limit: int = 1000,
    tables: Optional[tuple[str, ...]] = None,
    tempdir: Optional[str] = None,
    skip_versions: Optional[list[int]] = None,
):
    """
    Последовательное обновление БД ГАР на основе текущей версии БД и известных
    новых версий с сайта ФИАС.

    Args:
        limit: количество записей для пакетного сохранения. Не влияет на
            обновляемые объекты.
        tables: список импортируемых таблиц. Если передан None, импортируются
            все данные
        tempdir: путь до временной директории для загрузки и распаковки архивов
        skip_versions: номера версий обновлений, которые не будут выполняться и будут
            отмечены как пропущенные
    """
    last_processed_version = Version.objects.last_processed()

    if last_processed_version is None:
        raise TableListLoadingError('Not available. Please import the data before updating')

    skip_versions = set(skip_versions) if skip_versions is not None else set()

    versions = Version.objects.filter(ver__gt=last_processed_version.ver).order_by('ver')

    for version in versions:
        pre_update.send(sender=object.__class__, before=last_processed_version, after=version)

        if version.ver in skip_versions:
            update_version_status(version, VersionUpdateStatus.SKIPPED)

        elif version.delta_xml_url:
            try:
                update_data(
                    version=version,
                    limit=limit,
                    tables=tables,
                    tempdir=tempdir,
                )
            except ImporterError:
                update_version_status(version, VersionUpdateStatus.ERROR)

                raise
        else:
            # в версии нет ссылки на дельту
            if timezone.now().date() < (version.dumpdate + timedelta(hours=VERSION_WITHOUT_URL_TIMEOUT_HOURS)):
                # Время ожидания, когда ссылка для версии на сайте могла появиться ещё не прошло,
                # поэтому вызываем ошибку обновления
                update_version_status(version, VersionUpdateStatus.ERROR)

                raise VersionDoesNotHaveUrl(f'Версия {version} не имеет ссылки на файл с дельтой!')
            else:
                # Время ожидания, когда ссылка для версии на сайте могла появиться уже прошло,
                # поэтому больше не ждём и пропускаем версию
                update_version_status(version, VersionUpdateStatus.SKIPPED)

        post_update.send(sender=object.__class__, before=last_processed_version, after=version)

        last_processed_version = version


def get_reserve_delta_url(version):
    """
    Возвращает резервную ссылку для скачивания дельты
    """
    version_str = version.dumpdate.strftime('%Y.%m.%d')

    return f'https://file.nalog.ru/Downloads/{version_str}/gar_delta_xml.zip'



import datetime
from dataclasses import (
    dataclass,
)

import requests

from m3_gar.importer.exceptions import (
    DownloadFileInfoError,
)


@dataclass
class DownloadFileInfo:
    """Информация о файле с изменениями с сайта ФИАС."""

    version_id: int
    date: datetime.date
    gar_xml_full_url: str = ''
    gar_xml_delta_url: str = ''

    @classmethod
    def from_raw_info(cls, raw_info: dict) -> 'DownloadFileInfo':
        """Возвращает инстанс по информации скаченной с сайта ФИАС."""
        try:
            file_info = cls(
                version_id=raw_info['VersionId'],
                date=datetime.datetime.strptime(raw_info['Date'], '%d.%m.%Y').date(),
                gar_xml_full_url=raw_info['GarXMLFullURL'] or '',
                gar_xml_delta_url=raw_info['GarXMLDeltaURL'] or '',
            )
        except (KeyError, ValueError) as err:
            raise DownloadFileInfoError(
                f'Ошибка формата информации об обновлении с сайта ФИАС: {err}'
            )

        return file_info


class FIASClient:
    """Клиент к серверу ФИАС"""
    fias_source = 'https://fias.nalog.ru/WebServices/Public/'

    def get_all_download_file_info(self) -> list[DownloadFileInfo]:
        """Скачивает и возвращает перечень информации о файлах с обновлениями."""
        all_raw_info = requests.get(
            f'{self.fias_source}GetAllDownloadFileInfo',
        ).json()

        all_download_file_info = [
            DownloadFileInfo.from_raw_info(raw_info)
            for raw_info in all_raw_info
        ]

        return all_download_file_info


client = FIASClient()



import os
import tempfile
import zipfile
from urllib.error import (
    HTTPError,
    URLError,
)
from urllib.request import (
    urlretrieve,
)
from uuid import (
    uuid4,
)

from progress.bar import (
    Bar,
)

from m3_gar.importer.signals import (
    post_download,
    pre_download,
)
from m3_gar.importer.source.exceptions import (
    BadArchiveError,
    EmptyArchiveError,
    RetrieveError,
)
from m3_gar.importer.source.tablelist import (
    TableList,
)
from m3_gar.importer.source.wrapper import (
    LocalArchiveWrapper,
)


class ZipFile(zipfile.ZipFile):

    @classmethod
    def open_file(cls, source):
        try:
            return cls(source)
        except zipfile.BadZipfile:
            raise BadArchiveError(source)


class LocalArchiveTableList(TableList):
    wrapper_class = LocalArchiveWrapper

    def get_archive(self, source):
        if zipfile.is_zipfile(source):
            archive_class = ZipFile
        else:
            raise BadArchiveError(source)

        return archive_class.open_file(source)

    def load_data(self, source):
        archive = self.get_archive(source)

        if not archive.namelist():
            raise EmptyArchiveError(source)

        return self.wrapper_class(source=archive)


class DlProgressBar(Bar):
    message = 'Downloading: '
    suffix = '%(index)d/%(max)d. ETA: %(elapsed)s'
    hide_cursor = False


class RemoteArchiveTableList(LocalArchiveTableList):
    download_progress_class = DlProgressBar

    def load_data(self, source):
        progress = self.download_progress_class()

        def update_progress(count, block_size, total_size):
            progress.goto(int(count * block_size * 100 / total_size))

        if self.tempdir:
            tmp_file = os.path.join(self.tempdir, str(uuid4()))
        else:
            tmp_file = None

        pre_download.send(sender=self.__class__, url=source)
        try:
            path = urlretrieve(source, reporthook=update_progress, filename=tmp_file)[0]
        except (HTTPError, URLError) as e:
            raise RetrieveError(
                f'Can not download data archive at url `{source}`. '
                f'Error occurred: "{e}"'
            )
        progress.finish()
        post_download.send(sender=self.__class__, url=source, path=path)

        return super().load_data(source=path)
